#include <iostream>
#include <cuda.h>
#include <cuda_runtime.h>
#include <stdio.h>
#include <assert.h>
#include "cuda_check.cuh"

const int NUM_REPS = 1;

// Check errors
void postprocess(const double *ref, const double *res, int n, float ms)
{
  bool passed = true;
  for (int i = 0; i < n; i++)
    if (res[i] != ref[i]) {
      printf("%d %f %f\n", i, res[i], ref[i]);
      printf("%25s\n", "*** FAILED ***");
      passed = false;
      break;
    }
  if (passed)
    printf("%20.6f\n", ms/(1000.*NUM_REPS));
}


__global__ void transpose0(double *odata, const double *idata, int nx, int ny)
{
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  int j = blockIdx.y * blockDim.y + threadIdx.y;

  if ((i >= nx) || (j >= ny)) return;

  odata[j + i*ny] = idata[i + j*nx];
}

__global__ void dprint(double *d, int nx, int ny)
{
  int i,j;
  
  for (j=0; j<ny; j++)
  {
    printf("%4d:", j);
    for (i=0; i<nx; i++)
      printf(" %10.2g", d[i+j*nx]);
    printf("\n");
  }
  printf("\n");
}

void hprint(double *d, int nx, int ny)
{
  int i,j;
  
  for (j=0; j<ny; j++)
  {
    printf("%4d:", j);
    for (i=0; i<nx; i++)
      printf(" %10.2g", d[i+j*nx]);
    printf("\n");
  }
  printf("\n");
}

void vprint(double *d, int n)
{
  int i;
  
  for (i=0; i<n; i++)
    printf(" %10.2g", d[i]);
  printf("\n");
}


int main(int argc, char **argv)
{
  const int nx = (argc > 1) ? atol(argv[1]) : 1024;
  const int ny = nx;

  int blockDim = 16;

  dim3 dimGrid((nx+blockDim)/blockDim, (ny+blockDim)/blockDim, 1);
  dim3 dimBlock(blockDim, blockDim, 1);

  printf("Matrix size: %d %d, Block size: %d %d Grid size: %d %d\n", 
         nx, ny, dimBlock.x, dimBlock.y, dimGrid.x, dimGrid.y);
  
  double *h_idata = new double[nx*ny];
  double *h_tdata = new double[nx*ny];
  double *h_ref   = new double[nx*ny];

  // host
  for (int j = 0; j < ny; j++)
    for (int i = 0; i < nx; i++)
      h_idata[j*nx + i] = double(j*nx + i);

  if (nx < 8 && ny < 8) 
    hprint(h_idata, nx, ny);

  // pour verification
  for (int j = 0; j < ny; j++)
    for (int i = 0; i < nx; i++)
      h_ref[j*nx + i] = h_idata[i*nx + j];

  if (nx < 8 && ny < 8) 
    hprint(h_ref, nx, ny);
  
  const int mem_size = nx*ny*sizeof(double);
  double *d_idata, *d_tdata;

  CUDA_CHECK_OP( cudaMalloc(&d_idata, mem_size) );
  CUDA_CHECK_OP( cudaMalloc(&d_tdata, mem_size) );

  // device
  CUDA_CHECK_OP( cudaMemcpy(d_idata, h_idata, mem_size, cudaMemcpyHostToDevice) );
  cudaDeviceSynchronize();
  CUDA_CHECK_KERNEL();

  CUDA_CHECK_OP( cudaMemset(d_tdata, 0, mem_size) );
  cudaDeviceSynchronize();

  // events for timing
  cudaEvent_t startEvent, stopEvent;
  CUDA_CHECK_OP( cudaEventCreate(&startEvent) );
  CUDA_CHECK_OP( cudaEventCreate(&stopEvent) );

  // warmup
  transpose0<<<dimGrid, dimBlock>>>(d_tdata, d_idata, nx, ny);

  if (nx < 8 && ny < 8) 
    dprint<<<1,1>>>(d_tdata, nx, ny);

  cudaDeviceSynchronize();
  CUDA_CHECK_KERNEL();
  CUDA_CHECK_OP( cudaEventRecord(startEvent, 0) );

  for (int i = 0; i < NUM_REPS; i++)
     transpose0<<<dimGrid, dimBlock>>>(d_tdata, d_idata, nx, ny);
  cudaDeviceSynchronize();
  CUDA_CHECK_KERNEL();

  float ms = 0.0f;
  
  CUDA_CHECK_OP( cudaEventRecord(stopEvent, 0) );
  CUDA_CHECK_OP( cudaEventSynchronize(stopEvent) );
  CUDA_CHECK_OP( cudaEventElapsedTime(&ms, startEvent, stopEvent) );
  
  CUDA_CHECK_OP( cudaMemcpy(h_tdata, d_tdata, mem_size, cudaMemcpyDeviceToHost) );

  printf("\n%25s%25s\n", "Routine", "Execution time (s)");
  printf("%25s", "transpose0");
  postprocess(h_ref, h_tdata, nx * ny, ms);

  return 0;
}
